#![feature(proc_macro_hygiene, decl_macro)]

#[macro_use]
extern crate rocket;
#[macro_use]
extern crate rocket_contrib;
#[macro_use]
extern crate diesel;
#[macro_use]
extern crate serde_derive;
extern crate handlebars;
extern crate serde_json;

use std::collections::HashMap;

pub mod models;
pub mod schema;
use crate::diesel::GroupedBy;
use crate::models::{Klasse, Stunde, Vertretung};
use crate::schema::klassen;
use crate::schema::stunden;
use crate::schema::vertretungen;
use diesel::prelude::*;
use diesel::BelongingToDsl;
use diesel::QueryDsl;
use diesel::RunQueryDsl;
use rocket::request::{Form, FromParam};
use rocket::response::Redirect;
use rocket_contrib::databases;
use rocket_contrib::serve::StaticFiles;
use rocket_contrib::templates::Template;

struct StundenByID(i32);
#[derive(Debug)]
enum WebError {
    I32NotParsed,
}

impl<'a> FromParam<'_> for StundenByID {
    type Error = WebError;
    fn from_param(param: &rocket::http::RawStr) -> Result<Self, Self::Error> {
        if let Ok(n) = param.parse::<i32>() {
            Ok(StundenByID(n))
        } else {
            Err(WebError::I32NotParsed)
        }
    }
}

#[database("page_db")]
struct DbConn(databases::diesel::SqliteConnection);

#[derive(Debug, Serialize)]
struct ListAllContext {
    klassen: Vec<(Klasse, Vec<Vertretung>)>,
    parent: String,
}

fn query_all_classes(conn: DbConn) -> Vec<(Klasse, Vec<Vertretung>)> {
    let all_classes = klassen::table
        .load::<Klasse>(&*conn)
        .expect("database acess classes");
    let and_substitutions = Vertretung::belonging_to(&all_classes)
        .load::<Vertretung>(&*conn)
        .expect("database acess Vertretungen")
        .grouped_by(&all_classes);
    let combined_classes = all_classes
        .into_iter()
        .zip(and_substitutions)
        .filter(|x| x.1.len() > 0)
        .collect::<Vec<_>>();
    combined_classes
}

fn query_one_class<'a>(
    conn: DbConn,
    id: i32,
) -> (Klasse, HashMap<String, (Stunde, Vec<Vertretung>)>) {
    //println!("Suche Vertretungen:");
    let mut ret: HashMap<String, (Stunde, Vec<Vertretung>)> = HashMap::new();
    let class = klassen::dsl::klassen
        .filter(klassen::dsl::klassen_id.eq(id))
        .first::<Klasse>(&*conn)
        .expect("database acess classes");
    //println!("Für die Klassen: {:?}", class);
    //let stund: Vec<Stunde> = stunden::table.load(&*conn).expect("Loading Stunden");
    let times: Vec<(Vertretung, Stunde)> = Vertretung::belonging_to(&class)
        .inner_join(stunden::table.on(vertretungen::stunde_id.eq(stunden::stunden_id)))
        .load(&*conn)
        .expect("database access Stunden");
    /*println!(
        "Gefunden: {:?}",
        diesel::debug_query::<diesel::sqlite::Sqlite, _>(&times).to_string()
    );*/
    for (vert, stun) in times {
        //println!("{:?}, {:?}", vert, stun);
        let stun_ord = stun.ordinal.to_string();
        let verts = ret.entry(stun_ord).or_insert((stun, Vec::new()));
        verts.1.push(vert);
    }
    (class, ret)
}

#[get("/")]
fn index(conn: DbConn) -> Template {
    let combined_classes = query_all_classes(conn);
    Template::render(
        "uebersicht",
        ListAllContext {
            klassen: combined_classes,
            parent: "base".to_string(),
        },
    )
}

#[derive(Debug, Serialize)]
struct ListOneContext {
    klasse: Klasse,
    vertretungen: HashMap<String, (Stunde, Vec<Vertretung>)>,
    parent: String,
}

#[get("/klasse/<class_id>")]
fn klasse_view(conn: DbConn, class_id: i32) -> Template {
    let class = query_one_class(conn, class_id);
    let (klasse, vertretungen) = class;
    //println!("{:?}: {:?}", klasse, vertretungen);
    let cont = ListOneContext {
        klasse: klasse,
        vertretungen: vertretungen,
        parent: "base".to_string(),
    };
    Template::render("klasse", cont)
}

#[derive(Serialize)]
struct AddKlasseContext<'a> {
    klassen: Vec<Klasse>,
    parent: &'a str,
}

#[get("/add/klasse")]
fn get_klasse_form(conn: DbConn) -> Template {
    let klassen = klassen::table
        .order(klassen::ordnung.asc())
        .load::<Klasse>(&*conn)
        .expect("could not load stunden");
    Template::render(
        "add_klasse_form",
        AddKlasseContext {
            klassen,
            parent: "base",
        },
    )
}

#[get("/delete/klasse/<klassen_id>")]
fn delete_klasse(klassen_id: i32, conn: DbConn) -> Result<Redirect, String> {
    match diesel::delete(schema::klassen::table.filter(schema::klassen::klassen_id.eq(klassen_id)))
        .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/klasse")),
        Err(_) => Err("Klasse konnte nicht gelöscht werden!".to_string()),
    }
}

#[derive(FromForm, Insertable)]
#[table_name = "klassen"]
struct KlasseForm {
    stufe: i32,
    gruppe: String,
    titel: Option<String>,
    ordnung: i32,
}

#[post("/add/klasse", data = "<klasse>")]
fn post_klasse_form(klasse: Form<KlasseForm>, conn: DbConn) -> Result<Redirect, String> {
    match diesel::insert_into(schema::klassen::table)
        .values(klasse.into_inner())
        .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/klasse")),
        Err(_) => Err("Klasse konnte nicht erstellt werden!".to_string()),
    }
}

#[derive(Serialize)]
struct AddStundeContext<'a> {
    stunden: Vec<Stunde>,
    parent: &'a str,
}

#[get("/add/stunde")]
fn get_stunde_form(conn: DbConn) -> Template {
    let stunden = stunden::table
        .order(stunden::ordinal.asc())
        .load::<Stunde>(&*conn)
        .expect("could not load stunden");
    Template::render(
        "add_stunde_form",
        AddStundeContext {
            stunden,
            parent: "base",
        },
    )
}

#[get("/delete/stunde/<stunden_id>")]
fn delete_stunde(stunden_id: i32, conn: DbConn) -> Result<Redirect, String> {
    match diesel::delete(schema::stunden::table.filter(schema::stunden::stunden_id.eq(stunden_id)))
        .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/stunde")),
        Err(_) => Err("Stunde konnte nicht gelöscht werden!".to_string()),
    }
}

#[derive(FromForm, Insertable)]
#[table_name = "stunden"]
struct StundeForm {
    title: String,
    short: String,
    ordinal: i32,
}

#[post("/add/stunde", data = "<stunde>")]
fn post_stunde_form(stunde: Form<StundeForm>, conn: DbConn) -> Result<Redirect, String> {
    match diesel::insert_into(schema::stunden::table)
        .values(stunde.into_inner())
        .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/stunde")),
        Err(_) => Err("Stunde konnte nicht erstellt werden!".to_string()),
    }
}

#[derive(Serialize)]
struct AddVertretungContext<'a> {
    vertretungen: Vec<Vertretung>,
    parent: &'a str,
}

#[get("/add/vertretung")]
fn get_vertretung_form(conn: DbConn) -> Template {
    let loc_vertretungen = vertretungen::table
        .order(vertretungen::klasse_id.asc())
        .load::<Vertretung>(&*conn)
        .expect("could not load vertretungen");
    Template::render(
        "add_vertretung_form",
        AddVertretungContext {
            vertretungen: loc_vertretungen,
            parent: "base",
        },
    )
}

#[get("/delete/vertretung/<vertretungen_id>")]
fn delete_vertretung(vertretungen_id: i32, conn: DbConn) -> Result<Redirect, String> {
    match diesel::delete(
        schema::vertretungen::table
            .filter(schema::vertretungen::vertretungen_id.eq(vertretungen_id)),
    )
    .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/vertretung")),
        Err(_) => Err("Vertretung konnte nicht gelöscht werden!".to_string()),
    }
}

#[derive(FromForm, Insertable)]
#[table_name = "vertretungen"]
struct VertretungForm {
    klasse_id: i32,
    stunde_id: i32,
    fehlend: String,
    vertretung: String,
    kommentar: String,
}

#[post("/add/vertretung", data = "<vertretung>")]
fn post_vertretung_form(
    vertretung: Form<VertretungForm>,
    conn: DbConn,
) -> Result<Redirect, String> {
    match diesel::insert_into(schema::vertretungen::table)
        .values(vertretung.into_inner())
        .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/vertretung")),
        Err(_) => Err("Vertretung konnte nicht erstellt werden!".to_string()),
    }
}

fn mlen_helper_fun(
    h: &handlebars::Helper<'_, '_>,
    _: &handlebars::Handlebars,
    _: &handlebars::Context,
    _: &mut handlebars::RenderContext<'_>,
    out: &mut dyn handlebars::Output,
) -> handlebars::HelperResult {
    if let Some(param) = h.param(0) {
        if let serde_json::value::Value::Array(arr) = param.value() {
            let l = arr.len();
            let stunden = if l > 1 { "Stunden" } else { "Stunde" };
            out.write(&format!("{} {}", l, stunden))?;
            Ok(())
        } else {
            Err(handlebars::RenderError::new("Not a array"))
        }
    } else {
        Err(handlebars::RenderError::new(
            "No parameter given at least one required",
        ))
    }
}

fn main() {
    rocket::ignite()
        .attach(DbConn::fairing())
        .attach(Template::custom(|engines| {
            engines
                .handlebars
                .register_helper("length", Box::new(mlen_helper_fun));
        }))
        .mount(
            "/static",
            StaticFiles::from(concat!(env!("CARGO_MANIFEST_DIR"), "/static")),
        )
        .mount("/", routes![index])
        .mount("/", routes![klasse_view])
        .mount("/admin/", routes![get_klasse_form])
        .mount("/admin/", routes![post_klasse_form])
        .mount("/admin/", routes![get_stunde_form])
        .mount("/admin/", routes![post_stunde_form])
        .mount("/admin/", routes![get_vertretung_form])
        .mount("/admin/", routes![post_vertretung_form])
        .mount("/admin/", routes![delete_stunde])
        .mount("/admin/", routes![delete_klasse])
        .mount("/admin/", routes![delete_vertretung])
        .launch();
}
